import pytest
from loguru import logger
import shioaji as sj
from shioaji.constant import OrderState, StockOrderCond, Action

# from coverbot.deal import TFTDeal
from coverbot.cover import CoverBot, TFTDeal


deal = {
    "trade_id": "eb6055e4",
    "seqno": "425813",
    "ordno": "IE324",
    "exchange_seq": "430956",
    "broker_id": "9A9S",
    "account_id": "0291589",
    "action": "Buy",
    "code": "5871",
    "order_cond": "Cash",
    "order_lot": "Common",
    "price": 191.5,
    "quantity": 1,
    "web_id": "137",
    "custom_field": "",
    "ts": 1668482197,
    "name": "中租-KY",
    "time": "2022-11-15 11:16:37",
}
deal_sell = {
    "trade_id": "eb6055e4",
    "seqno": "425813",
    "ordno": "IE324",
    "exchange_seq": "430956",
    "broker_id": "9A9S",
    "account_id": "0291589",
    "action": "Sell",
    "code": "5871",
    "order_cond": "Cash",
    "order_lot": "Common",
    "price": 191.5,
    "quantity": 1,
    "web_id": "137",
    "custom_field": "",
    "ts": 1668482197,
    "name": "中租-KY",
    "time": "2022-11-15 11:16:37",
}


def test_sell_deal_action(sut: CoverBot):
    sut.deal_action(deal_sell)
    sut.show()
    assert sut.deals[deal_sell["code"]].quantity == -1


def test_deal_action(sut: CoverBot):
    assert sut.deals.get(deal["code"], None) == None
    sut.deal_action(deal)
    assert sut.deals[deal["code"]].quantity == 1
    sut.deal_action(deal_sell)
    assert sut.deals[deal["code"]].quantity == 0
    sut.deal_action(deal_sell)
    assert sut.deals[deal["code"]].quantity == -1


def test_deal_date(sut: CoverBot):
    sut.order_handler(OrderState.TFTDeal, deal)
    assert sut.deals[deal["code"]].quantity == 1


def test_show(sut, deal_data):
    for d in deal_data:
        if d["action"] == "Sell":
            sut.order_handler(OrderState.TFTDeal, d)
    logger.info(sut.show())



from loguru import logger
import pytest
import shioaji as sj
import pickle
from coverbot.cover import CoverBot
import pathlib
from typing import Dict, List

contract_filename = f"{pathlib.Path(__file__).parent.absolute()}/data/contracts.pkl"


@pytest.fixture
def deal_data() -> List[Dict]:
    deal_filename = f"{pathlib.Path(__file__).parent.absolute()}/data/_deals.pkl"
    with open(deal_filename, "rb") as f:
        test_data = pickle.load(f)
    return test_data


@pytest.fixture
def mock_api() -> sj.Shioaji:
    api: sj.Shioaji = sj.Shioaji()
    with open(contract_filename, "rb") as f:
        api.Contracts = pickle.load(f)
    return api


@pytest.fixture
def sut(mock_api: sj.Shioaji) -> CoverBot:
    sut = CoverBot(mock_api)
    return sut



from typing import Dict, List, Optional
from dataclasses import dataclass, field, fields
from shioaji.constant import OrderState, Action, StockOrderCond
from shioaji.contracts import Contract


@dataclass
class TFTDeal:
    code: str
    action: Action
    price: float
    quantity: int
    order_cond: StockOrderCond
    web_id: str
    custom_field: str
    time: str

    def __init__(self, **kwargs):
        names = set([f.name for f in fields(self)])
        for k, v in kwargs.items():
            if k in names:
                setattr(self, k, v)


class Deal:
    def __init__(self, contract: Contract):
        self.contract: Contract = contract
        self.buy_quantity: int = 0
        self.sell_quantity: int = 0
        self.amount: float = 0
        self.stop_price: float = 0.0
        self.first_action: Optional[Action] = None
        self.entry_price: float = 0.0
        self.stop_loss_pct: float = 0.09
        self.buy_collection: List[TFTDeal] = []
        self.sell_collection: List[TFTDeal] = []

    @property
    def quantity(self):
        return self.buy_quantity - self.sell_quantity

    # def __repr__(self) -> str:
    #     return f"Code : {self.contract}\n quantity:{self.quantity} buy: {len(self.buy_collection)} sell: {len(self.sell_collection)}"
    def to_dict(self):
        d = {
            "contract": f"{self.contract.code} {self.contract.name}",
            "action": self.first_action,
            "quantity": self.quantity,
            "entry_price": self.entry_price,
            "buy_records": len(self.buy_collection),
            "sell_records": len(self.sell_collection),
        }
        return d

    # def __dict__(self):

    def __str__(self) -> str:
        return str(
            {
                "contract": f"{self.contract.code} {self.contract.name}",
                "action": self.first_action,
                "quantity": self.quantity,
                "entry_price": self.entry_price,
                "buy_records": len(self.buy_collection),
                "sell_records": len(self.sell_collection),
            }
        )

    def apply(self, tftdeal: Dict):
        deal = TFTDeal(**tftdeal)
        amount = deal.quantity * deal.price * 1000
        if not self.first_action:
            self.first_action = deal.action

        if deal.action == Action.Buy:
            self.buy_quantity += deal.quantity
            self.buy_collection.append(deal)

        elif deal.action == Action.Sell:
            self.sell_quantity += deal.quantity
            self.sell_collection.append(deal)

        if self.first_action == deal.action:
            self.amount += amount
            quantity = (
                self.buy_quantity
                if self.first_action == Action.Buy
                else self.sell_quantity
            )
            self.entry_price = self.amount / quantity / 1000.0



from loguru import logger
from typing import Dict, List, Any, Union
from dataclasses import dataclass, field, fields
import shioaji as sj
from shioaji.contracts import Contract

from shioaji.constant import OrderState, Action, StockOrderCond
from .deal import TFTDeal, Deal


class CoverBot:
    def __init__(self, api: sj.Shioaji):
        self.api = api
        self.api.set_order_callback(self.order_handler)
        self.deals: Dict[str, Deal] = {}

    def set_stop_loss_pct(self, code: str):
        contract = self.api.Contracts.Stocks[code]
        deal = self.deals[code] = self.deals.get(code, Deal(contract))

    def show(self):
        return [v.to_dict() for _, v in self.deals.items()]

    def deal_action(self, tftdeal: Dict[str, Any]) -> None:
        code = tftdeal.get("code", "")
        contract = self.api.Contracts.Stocks[code]
        if not contract:
            logger.warning(f"[{code}] not exist.")
        else:
            deal = self.deals[code] = self.deals.get(code, Deal(contract))
            deal.apply(tftdeal)

    def order_handler(self, order_state: OrderState, msg: Dict) -> None:
        if order_state == OrderState.TFTDeal:
            self.deal_action(msg)
